/// <summary>
/// ****************************************************************************
/// @FILE:      AIOUSB_API.cs
///
/// @BRIEF:     C# wrapper for AIOUSB.DLL 
///
/// @HISTORY:   2016-03-25  J.Bendor    Created from existing wrapper
/// 
/// @COPYRIGHT: ACCES-IO, All Rights Reserved.
/// ****************************************************************************
/// </summary>

using System;
using System.Runtime.InteropServices;


namespace AIOUSB_API
{
    public struct TUSBAI1616Config    // based on USB-Software-Reference (pdf)
    {
        public byte[] ChannelsRange;
        public byte       CalibMode;
        public byte        TrigMode;
        public byte     StartStopCh;
        public byte      Oversample;
        public byte   StartStopChEx;

        public static byte[] ToBytes(TUSBAI1616Config aConfig)
        {
            byte[] bytes_array = new byte[aConfig.ChannelsRange.Length + 5];

            aConfig.ChannelsRange.CopyTo(bytes_array, 0);

            int offset = aConfig.ChannelsRange.Length - 1;

            bytes_array[ ++offset ] = aConfig.CalibMode;
            bytes_array[ ++offset ] = aConfig.TrigMode;
            bytes_array[ ++offset ] = aConfig.StartStopCh;
            bytes_array[ ++offset ] = aConfig.Oversample;
            bytes_array[ ++offset ] = aConfig.StartStopChEx;

            return bytes_array;
        }

        public TUSBAI1616Config(byte trigMode, byte calibMode, byte overMode, byte[] ranges = null)
        {
            int num_channels = (ranges == null) ? (int) AIOUSB.MAX_CFG_CHANNELS : ranges.Length;

            ChannelsRange = new byte[num_channels];

            // Set the A/D range to 0-10V, which will become 0-16V on this DPK.
            for (int idx = 0; idx < ChannelsRange.Length; ++idx)
            {
                ChannelsRange[idx] = (ranges == null) ? (byte) 0 : ranges[idx];
            }
            
            TrigMode = trigMode;    // selected channels, each counter rising edge. Leave counter 0 external.
            CalibMode = calibMode;  // Take actual data, not internal calibration sources.
            Oversample = overMode;  // No oversample.
            StartStopCh = 0;
            StartStopChEx = 0;
        }

    } // ends struct TUSBAI1616Config


    public sealed class AIOUSB      // interface faciliated by "AIOUSB.DLL"
    {
        public const uint MAX_NUM_DEVICES  = 7;
        public const uint MAX_NUM_CHANNELS = 64;
        public const uint MAX_CFG_CHANNELS = 16;
        public const uint MAX_DEV_NAME_LNG = 99;

        public const uint ERR_SUCCESS = 0;
        public const uint ERR_FAILURE = 1;

        // special constants for DeviceIndex (aka DI)
        public static uint diNone  { get { return 0xFFFFFFFF; } }  // -1: invalid DI
        public static uint diFirst { get { return 0xFFFFFFFE; } }  // -2: 1ST board, lowest DI
        public static uint diOnly  { get { return 0xFFFFFFFD; } }  // -3: One and only board 


        // ADCallback is called by a thread spawned by the driver --- pData is an array short[]
        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        public delegate void ADCallback(IntPtr pData, uint NumDataBytes, uint Flags, uint Context);


        private AIOUSB() { }   // private c'tor prevents instantiation

        public static uint ClearDevices()
        {
            return DLL_API.ClearDevices();
        }

        public static uint GetDevices()
        {
            return DLL_API.GetDevices();
        }

        public static uint GetDeviceSerialNumber(uint uDI, out ulong SerNum)
        {
            return DLL_API.GetDeviceSerialNumber(uDI, out SerNum);
        }

        public static uint AIOUSB_SetStreamingBlockSize(uint uDI, uint blockSize)
        {
            return DLL_API.AIOUSB_SetStreamingBlockSize(uDI, blockSize);
        }

        public static uint ADC_QueryCal(uint uDI)
        {
            return DLL_API.ADC_QueryCal(uDI);
        }

        public static uint ADC_BulkPoll(uint uDI, out uint numBytesLeft)
        {
            return DLL_API.ADC_BulkPoll(uDI, out numBytesLeft);
        }

        public static uint ADC_BulkAcquire(uint uDI, uint numBytesLeft, ushort[] pDataAD)
        {
            return DLL_API.ADC_BulkAcquire(uDI, numBytesLeft, pDataAD);
        }

        public static uint QueryDeviceInfo(uint uDI, out uint uPID, out string strDevName)
        {
            uint uCounters = 0;
            uint uDIObytes = 0;
            uint uName_Lng = 0;
            return QueryDeviceInfo(uDI, out uPID, ref uName_Lng, out strDevName, out uDIObytes, out uCounters);
        }

        public static uint QueryDeviceInfo(uint uDI, out uint uPID, ref uint uNameLng, out string strDevName, out uint uDIObytes, out uint uCounters)
        {
            strDevName = "";

            uNameLng = MAX_DEV_NAME_LNG;

            byte[] a_ascii = new byte[uNameLng + 1];

            uint err_code = DLL_API.QueryDeviceInfo(uDI, out uPID, ref uNameLng, a_ascii, out uDIObytes, out uCounters);

            if (err_code == ERR_SUCCESS)
            {
                strDevName = System.Text.Encoding.UTF8.GetString(a_ascii, 0, (int)uNameLng);
            }

            return err_code;
        }

        public static uint ADC_SetConfig(uint uDI, TUSBAI1616Config aConfigBuf)
        {
            byte[] cfg_buf = TUSBAI1616Config.ToBytes(aConfigBuf);

            uint   cfg_lng = (uint)cfg_buf.Length;

            return ADC_SetConfig(uDI, cfg_buf, ref cfg_lng);
        }

        public static uint ADC_SetConfig(uint uDI, byte[] configData, ref uint configSize)
        {
            return DLL_API.ADC_SetConfig(uDI, configData, ref configSize);
        }

        public static uint ADC_SetCal(uint uDI, string configFileName)
        {
            int   config_name_length = configFileName.Length;

            char[] config_name_chars = new char[config_name_length + 1];

            configFileName.CopyTo(0, config_name_chars, 0, config_name_length);

            return DLL_API.ADC_SetCal(uDI, config_name_chars);
        }

        public static uint CTR_StartOutputFreq(uint uDI, uint uBlockIndex, ref double HzFreq)
        {
            return DLL_API.CTR_StartOutputFreq(uDI, uBlockIndex, ref HzFreq);
        }

        public static uint ADC_BulkContinuousCallbackStartClocked(uint uDI, uint uBufSize, uint uBaseBufCount, uint uContext, ADCallback callback, ref double pHz)
        {
            return DLL_API.ADC_BulkContinuousCallbackStartClocked(uDI, uBufSize, uBaseBufCount, uContext, callback, ref pHz);
        }

        public static uint ADC_BulkContinuousCallbackStart(uint uDI, uint uBufSize, uint uBaseBufCount, uint uContext, ADCallback callback)
        {
            return DLL_API.ADC_BulkContinuousCallbackStart(uDI, uBufSize, uBaseBufCount, uContext, callback);
        }

        public static uint ADC_BulkContinuousEnd(uint uDI, out uint aIOStatus)
        {
            return DLL_API.ADC_BulkContinuousEnd(uDI, out aIOStatus);
        }

    } // ends class AIOUSB


    public sealed class DLL_API     // simple wrapper for "AIOUSB.DLL"
    {
        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ClearDevices();

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint GetDevices();

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_QueryCal(uint DeviceIndex);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_BulkPoll(uint DeviceIndex, out uint numBytesLeft);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_BulkAcquire(uint DeviceIndex, uint numBytesLeft, ushort[] pDataAD);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint AIOUSB_SetStreamingBlockSize(uint DeviceIndex, uint blockSize);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint GetDeviceSerialNumber(uint DeviceIndex, out ulong SerNum);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint QueryDeviceInfo(uint DeviceIndex, out uint pPID, ref uint pNameSize, [In, Out] byte[] pName, out uint pDIOBytes, out uint pCounters);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_SetConfig(uint DeviceIndex, [In] byte[] pConfigBuf, ref uint pConfigBufSize);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_SetCal(uint DeviceIndex, [In] Char[] pCalFileName);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_SetCalAndSave(uint DeviceIndex, [In] Char[] pCalFileName, [In] Char[] pSaveCalFileName_optional);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint CTR_StartOutputFreq(uint DeviceIndex, uint BlockIndex, ref double Hz);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_BulkContinuousCallbackStartClocked(uint DeviceIndex, uint BufSize, uint BaseBufCount, uint Context, AIOUSB.ADCallback pCallback, ref double pHz);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_BulkContinuousCallbackStart(uint DeviceIndex, uint BufSize, uint BaseBufCount, uint Context, AIOUSB.ADCallback pCallback);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl)]
        public static extern uint ADC_BulkContinuousEnd(uint DeviceIndex, out uint pIOStatus);

        private DLL_API() { }   // private c'tor prevents instantiation

    } // ends class DLL_API

} // ends namespace AIOUSB_API
