/// <summary>
/// ****************************************************************************
/// @FILE:      MainForm.cs
///
/// @BRIEF:     implements the MainForm window for the ContLog app.
///
/// @HISTORY:   2016-03-25  J.Bendor    Created from "WorkUnit.pas"
/// 
/// @COPYRIGHT: ACCES-IO, All Rights Reserved.
/// ****************************************************************************
/// </summary>

using System;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;

using AIOUSB_API;
using DetectForm;


namespace MainForm
{
    public partial class TMainForm : Form
    {
        public const uint WM_USER = 0x0400;
        public const uint ASYNC_RCVD_SAMPLES = WM_USER + 0x100;
        public const uint ASYNC_LOST_SAMPLES = WM_USER + 0x101;
        public const uint ASYNC_ENDED_STREAM = WM_USER + 0x102;
        public const uint ASYNC_ABORT_STREAM = WM_USER + 0x103;

        private AIOUSB.ADCallback A2D_callback;

        // thread-safe delegate used for GUI updates triggered by non-GUI threads
        public delegate void Safe_GUI_Updater(object[] eventArgs);

        public TMainForm()
        {
            InitializeComponent();
        }

        public void TMainForm_FormLoad(System.Object Sender, System.EventArgs evarg)
        {
            string str_cpu = (Marshal.SizeOf(IntPtr.Zero) > 4) ? "64 bit" : "32 bit";

            do_append_to_report("Started App ContLog (" + str_cpu + ") \r\n");

            A2D_callback = new AIOUSB.ADCallback(WorkUnit.AD_Callback);

            WorkUnit.mMainForm = this;

            this.Show();

            btnExit.Enabled = true;
            btnStop.Enabled = false;
            btnDetect.Enabled = true;
            btnCapture.Enabled = do_accept_device(AIOUSB.diOnly);

            // possibly --- let the user select a device
            if (!btnCapture.Enabled)
            {
                btnDetect_Click(null, null);
            }
        }

        public void TMainForm_FormClosing(System.Object Sender, System.EventArgs evarg)
        {
            InvokeAsyncEvent(ASYNC_ABORT_STREAM);
        }

        private void btnDetect_Click(object sender, EventArgs e)
        {
            TDetectForm detect_form = new TDetectForm();

            if (detect_form.ShowDialog() == DialogResult.OK)
            {
                btnCapture.Enabled = do_accept_device(detect_form.DeviceIndex);
            }
            else
            {
                btnCapture.Enabled = do_accept_device(AIOUSB.diNone);
            }
        }

        private void btnExit_Click(object sender, EventArgs e)
        {
            InvokeAsyncEvent(ASYNC_ABORT_STREAM);
            this.Close();
        }

        public void btnStop_Click(System.Object Sender, System.EventArgs _e1)
        {
            InvokeAsyncEvent(ASYNC_ABORT_STREAM);
        }

        public void btnCapture_Click(System.Object Sender, System.EventArgs evarg)
        {
            do_start_data_capture();
        }

        public bool InvokeAsyncEvent(uint eventType, uint eventData1 = 0, uint eventData2 = 0)
        {
            Delegate the_method = new Safe_GUI_Updater(handle_async_GUI_update_event);

            object event_args = new object[] { eventType, eventData1, eventData2 };

            this.BeginInvoke(the_method, event_args);

            return true;
        }

        private void handle_async_GUI_update_event(object[] eventArgs)
        {
            uint event_type = (uint)eventArgs[0];

            if (event_type == ASYNC_RCVD_SAMPLES)
            {
                uint num_samples = (uint)eventArgs[1];
                uint scans_count = (uint)eventArgs[2];
                do_append_to_report(string.Format("Samples={0}  Scans={1}", num_samples, scans_count));
                return;
            }

            if (event_type == ASYNC_LOST_SAMPLES)
            {
                uint lost_length = (uint)eventArgs[1];
                do_append_to_report(string.Format("Lost ({0}) Bytes", lost_length));
                return;
            }

            if (event_type == ASYNC_ENDED_STREAM)
            {
                do_append_to_report("Capture Ended \r\n");
                return;
            }

            if (event_type == ASYNC_ABORT_STREAM)
            {
                do_stop_data_capture();
                return;
            }

            return;
        }

        private bool do_append_to_report(string strLine, bool isNewLine = true)
        {
            if (isNewLine)
            {
                ReportMemo.AppendText("\r\n");
                ReportMemo.AppendText(DateTime.Now.ToString("HH.mm.ss.fff : "));
            }
            ReportMemo.AppendText(strLine);
            return isNewLine;
        }

        private bool do_accept_device(uint deviceIndex)
        {
            WorkUnit.mDIX = deviceIndex;

            uint uPID = 0;
            string Name = "";

            // If we only have the one board, QueryDeviceInfo() with diOnly will succeed
            // with the right PID. If we have more than one board, or haven't plugged it
            // in, it will fail, the detector form allows the user to pick a device.
            uint status = AIOUSB.QueryDeviceInfo(WorkUnit.mDIX, out uPID, out Name);

            bool ok_1 = ((uPID >= 0x804F) && (uPID <= 0x8053));
            bool ok_2 = ((uPID >= 0x814F) && (uPID <= 0x8153));

            if ((status != AIOUSB.ERR_SUCCESS) || (!(ok_1 || ok_2)))
            {
                return false;
            }

            status = AIOUSB.ADC_SetConfig(WorkUnit.mDIX, WorkUnit.mConfig);

            if (status != AIOUSB.ERR_SUCCESS)
            {
                string info = string.Format("ADC_SetConfig({0:X},_) \r\n\n", WorkUnit.mDIX);
                do_append_to_report("Error=" + status.ToString() + " --- FAILED! " + info);
            }
            else
            {
                string info = string.Format("ADC_SetConfig({0:X},_) \r\n\n", WorkUnit.mDIX);
                do_append_to_report("OK!   " + info);
            }

            return (status == AIOUSB.ERR_SUCCESS);
        }

        private bool do_stop_data_capture()
        {
            if (!btnStop.Enabled)
            {
                return false;
            }

            // close the log file --- also prevents processing arriving samples
            WorkUnit.CloseLog();

            do_append_to_report("Stopping...");

            uint IOStatus = 0;
            uint EndStatus = AIOUSB.ADC_BulkContinuousEnd(WorkUnit.mDIX, out IOStatus);

            // Make sure we got the "stream ended" message first.
            if ((EndStatus == AIOUSB.ERR_SUCCESS) && (IOStatus == AIOUSB.ERR_SUCCESS))
            {
                do_append_to_report("Acquisition ended OK!");
            }
            else if (EndStatus == AIOUSB.ERR_SUCCESS)
            {
                do_append_to_report("Acquisition ended with IOStatus error " + IOStatus.ToString());
            }
            else
            {
                do_append_to_report("Ending acquisition with EndStatus error " + EndStatus.ToString());
            }

            btnStop.Enabled = false;
            btnDetect.Enabled = true;
            btnCapture.Enabled = true;

            btnCapture.Focus();

            return true;
        }

        private bool do_start_data_capture()
        {
            if (!btnCapture.Enabled)
            {
                return false;
            }

            uint   uPID = 0;
            string Name = "";

            uint status = AIOUSB.QueryDeviceInfo(WorkUnit.mDIX, out uPID, out Name);

            if (status != AIOUSB.ERR_SUCCESS)
            {
                string info = string.Format("QueryDeviceInfo({0:X},_,_) \r\n\n", WorkUnit.mDIX);
                do_append_to_report("Error=" + status.ToString() + " --- FAILED! " + info);
                return false;
            }

            double Hz = 0.0;    // Stop the counter, in case it was running.
            status = AIOUSB.CTR_StartOutputFreq(WorkUnit.mDIX, 0, ref Hz);

            if (status != AIOUSB.ERR_SUCCESS)
            {
                string info = string.Format("CTR_StartOutputFreq({0:X},0,Hz) \r\n\n", WorkUnit.mDIX);
                do_append_to_report("Error=" + status.ToString() + " --- FAILED! " + info);
                return false;
            }

            WorkUnit.mNumChannels = (int)AIOUSB.MAX_NUM_CHANNELS;  // Select all channels.
            byte EndChannel = (byte)(WorkUnit.mNumChannels - 1);
            byte StartChannel = 0;

            WorkUnit.mConfig.StartStopCh = (byte)((EndChannel << 4) | (StartChannel & 0xF));
            WorkUnit.mConfig.StartStopChEx = (byte)((EndChannel & 0xF0) | (StartChannel >> 4));

            status = AIOUSB.ADC_SetConfig(WorkUnit.mDIX, WorkUnit.mConfig);

            if (status != AIOUSB.ERR_SUCCESS)
            {
                string info = string.Format("ADC_SetConfig({0:X},_) \r\n\n", WorkUnit.mDIX);
                do_append_to_report("Error=" + status.ToString() + " --- FAILED! " + info);
                return false;
            }

            string calib_name = ":AUTO:";
            status = AIOUSB.ADC_SetCal(WorkUnit.mDIX, calib_name);

            // reprt and ignore error if the device doesn't support calibration
            if (status != AIOUSB.ERR_SUCCESS)
            {
                string info = string.Format("ADC_SetCal({0:X},\"{1}\") \r\n\n", WorkUnit.mDIX, calib_name);
                string code = string.Format("Error=0x{0:X} ---- ", status);
                do_append_to_report(code + info);
            }

            // when caps-lock is pressed the log has some extra info
            WorkUnit.mIsLogDBG = Control.IsKeyLocked(Keys.CapsLock);
            string csv_suffix = WorkUnit.mIsLogDBG ? ".CSV" : ".csv";

            string str_time_now = DateTime.Now.ToString("yyyyMMdd_HHmmss");
            string str_csv_name = string.Format("PID={0:X4}_{1}{2}", uPID, str_time_now, csv_suffix);

            WorkUnit.mLogPath = Path.Combine(Environment.CurrentDirectory, str_csv_name);
            WorkUnit.mLogFile = new FileStream(WorkUnit.mLogPath, FileMode.Create);
            WorkUnit.CloseLog();

            uint mode = ((uint)FileAccess.ReadWrite | (uint)FileShare.Read);
            WorkUnit.mLogFile = new FileStream(WorkUnit.mLogPath, (FileMode)mode);

            WorkUnit.mLogText.Clear();
            WorkUnit.mLogText.Append("SCAN#,");
            for (uint idx = 0; idx < AIOUSB.MAX_NUM_CHANNELS; ++idx)
            {
                WorkUnit.mLogText.Append(string.Format("ch{0:D02},", idx));
            }
            WorkUnit.mLogText.Append(string.Format("_\r\n"));

            WorkUnit.WriteLog(WorkUnit.mLogText.ToString(), true);
            WorkUnit.mLogText.Clear();

            WorkUnit.mNextChannel = 0;
            WorkUnit.mNumSamples = 0;
            WorkUnit.mNumChunks = 0;
            WorkUnit.mNumScans = 1;

            uint callback_buf_lng = WorkUnit.NUM_SAMPLES_PER_CHUNK * 2;
            uint callback_context = 0;
            status = AIOUSB.ADC_BulkContinuousCallbackStart(WorkUnit.mDIX,
                                                            callback_buf_lng,
                                                            64,
                                                            callback_context,
                                                            A2D_callback);

            if (status != AIOUSB.ERR_SUCCESS)
            {
                string info = string.Format("ADC_BulkContinuousCallbackStart({0:X},,,) \r\n\n", WorkUnit.mDIX);
                do_append_to_report("Error=" + status.ToString() + " --- FAILED! " + info);
                WorkUnit.CloseLog();
                return false;
            }

            // Since we've put it in scan mode, this is per frequency channel.
            Hz = WorkUnit.SAMPLING_FREQUENCY_HZ / AIOUSB.MAX_NUM_CHANNELS;
            status = AIOUSB.CTR_StartOutputFreq(WorkUnit.mDIX, 0, ref Hz);

            if (status != AIOUSB.ERR_SUCCESS)
            {
                string info = string.Format("CTR_StartOutputFreq({0:X},,) \r\n\n", WorkUnit.mDIX);
                do_append_to_report("Error=" + status.ToString() + " --- FAILED! " + info);
                WorkUnit.CloseLog();
                return false;
            }

            do_append_to_report(string.Format("Started LOG ({0})", WorkUnit.mLogPath));

            btnCapture.Enabled = false;
            btnDetect.Enabled = false;
            btnStop.Enabled = true;

            btnStop.Focus();

            return true;
        }

    } // ends class TMainForm


    public sealed class WorkUnit
    {
        // note: CPU speed limits feed rate (sampling_frequency x chunk_size)
        // note: The app could freeze if the CPU is too slow for the feed rate.
        // note: The DLL does not provide a throttle to "adjust" its feed rate.
        // note: The app's "Release" version runs much faster than the "Debug".

        public const double   SAMPLING_FREQUENCY_HZ = (250.0 * 1000);
        public const int      NUM_SAMPLES_PER_CHUNK = 1024;
        public const int      CHUNKS_PER_GUI_UPDATE = 100;

        public static TMainForm           mMainForm = null;
        public static TUSBAI1616Config      mConfig = new TUSBAI1616Config(0x15, 0, 0);
        public static FileStream           mLogFile = null;
        public static StringBuilder        mLogText = new StringBuilder(NUM_SAMPLES_PER_CHUNK*5);
        public static string               mLogPath = "";
        public static bool                mIsLogDBG = false;
        public static short[]             mDataRcvd = new short[NUM_SAMPLES_PER_CHUNK + 1];
        public static uint                     mDIX = 0;
        public static uint             mNextChannel = 0;
        public static uint             mNumChannels = 0;
        public static uint              mNumSamples = 0;
        public static uint               mNumChunks = 0;
        public static uint                mNumScans = 0;


        private WorkUnit() { }  // private c'tor prevents instantiation

        // safely write string to log file
        public static int WriteLog(string strLogText, bool flushToDisk = false)
        {
            lock (WorkUnit.mMainForm)
            {
                if (WorkUnit.mLogFile != null)
                {
                    byte[] ascii_text = Encoding.ASCII.GetBytes(strLogText);

                    WorkUnit.mLogFile.Write(ascii_text, 0, ascii_text.Length);

                    if (flushToDisk)
                    {
                        WorkUnit.mLogFile.Flush(true);
                    }
                }
            }

            return strLogText.Length;
        }

        // safely close the log file
        public static long CloseLog()
        {
            FileStream the_log_file = null;

            lock(WorkUnit.mMainForm)
            {
                the_log_file = WorkUnit.mLogFile;
                WorkUnit.mLogFile = null;
            }

            if (the_log_file == null)
            {
                return -1;
            }

            the_log_file.Flush(true);

            long log_file_size = the_log_file.Length;

            the_log_file.Close();

            return log_file_size;
        }

        // async callback called from thread in the device's DLL
        public static void AD_Callback(IntPtr pData, uint NumDataBytes, uint Flags, uint Context)
        {
            try
            {
                WorkUnit.do_receive_async_data(pData, NumDataBytes, Flags, Context);
            }
            catch (Exception ex)
            {
                WorkUnit.mLogText.Append("\r\n" + "EXCEPTION \r\n" + ex.Message + "\r\n");

                WorkUnit.WriteLog( WorkUnit.mLogText.ToString() );

                WorkUnit.CloseLog();

                WorkUnit.mMainForm.InvokeAsyncEvent(TMainForm.ASYNC_ENDED_STREAM);
            }
        }

        // receive and log data captured by the device's DLL
        private static void do_receive_async_data(IntPtr pData, uint NumDataBytes, uint Flags, uint Context)
        {
            // possibly --- end of stream
            if ((Flags & 0x2) != 0)
            {
                WorkUnit.CloseLog();
                WorkUnit.mMainForm.InvokeAsyncEvent(TMainForm.ASYNC_ENDED_STREAM);
                return;
            }

            // possibly --- log file closed
            if (WorkUnit.mLogFile == null)
            {
                return;
            }

            // possibly --- periodically flush to disk and update the GUI
            bool flush_to_disk = ((++WorkUnit.mNumChunks % CHUNKS_PER_GUI_UPDATE) == 1);
            if (flush_to_disk)
            {
                WorkUnit.mMainForm.InvokeAsyncEvent(TMainForm.ASYNC_RCVD_SAMPLES, mNumSamples, WorkUnit.mNumScans);
            }

            if (WorkUnit.mIsLogDBG)
            {
                string str_info = string.Format("\r\n{0}  [{1}]  #{2} \r\n", 
                                                DateTime.Now.ToString("HH.mm.ss.fff"), 
                                                NumDataBytes, 
                                                WorkUnit.mNumChunks);

                WorkUnit.WriteLog( string.Format(str_info) );
            }

            // possibly --- invalid data length
            uint num_data_items = (NumDataBytes / 2);
            if ( (num_data_items < 1) || (num_data_items > WorkUnit.mDataRcvd.Length) )
            {
                WorkUnit.mMainForm.InvokeAsyncEvent(TMainForm.ASYNC_LOST_SAMPLES, NumDataBytes);
                return;
            }

            Marshal.Copy(pData, WorkUnit.mDataRcvd, 0, (int)num_data_items);

            for (uint index = 0;  index < num_data_items;  ++index)
            {
                if (WorkUnit.mLogText.Length == 0)
                {
                    WorkUnit.mLogText.Append(string.Format("{0}#,", WorkUnit.mNumScans));
                }

                uint sample = (uint) WorkUnit.mDataRcvd[index];

                WorkUnit.mLogText.Append( string.Format("{0:X},", sample) );

                if (++WorkUnit.mNextChannel < WorkUnit.mNumChannels)
                {
                    continue;
                }

                WorkUnit.mLogText.Append(string.Format("_\r\n"));

                WorkUnit.WriteLog(WorkUnit.mLogText.ToString(), flush_to_disk );

                WorkUnit.mNumScans += 1;

                WorkUnit.mNextChannel = 0;

                WorkUnit.mLogText.Clear();
            }

            mNumSamples += num_data_items;

            return;
        }

    } // ends class WorkUnit

}   // ends namespace MainForm
